@@ -2,6 +2,12 @@
 // SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #include "testserver.h"
+
+#include "generated/namespace_qtopcuatestmodel_generated.h"
+#include "generated/types_qtopcuatestmodel_generated.h"
+#include "generated/types_qtopcuatestmodel_generated_handling.h"
+#include "generated/qtopcuatestmodel_nodeids.h"
+
 #include "qopen62541utils.h"
 #include "qopen62541valueconverter.h"
 #include <QtOpcUa/qopcuatype.h>
@@ -68,7 +74,7 @@ static UA_ByteString loadFile(const QString &filePath) {
         return fileContents;
 
     if (file.read(reinterpret_cast<char*>(fileContents.data), fileContents.length) != static_cast<qint64>(fileContents.length)) {
-        UA_ByteString_deleteMembers(&fileContents);
+        UA_ByteString_clear(&fileContents);
         fileContents.length = 0;
         return fileContents;
     }
@@ -81,9 +87,9 @@ bool TestServer::createSecureServerConfig(UA_ServerConfig *config)
     const QString privateKeyFilePath = QLatin1String(":/pki/own/private/open62541-testserver.der");
 
     UA_ByteString certificate = loadFile(certificateFilePath);
-    UaDeleter<UA_ByteString> certificateDeleter(&certificate, UA_ByteString_deleteMembers);
+    UaDeleter<UA_ByteString> certificateDeleter(&certificate, UA_ByteString_clear);
     UA_ByteString privateKey = loadFile(privateKeyFilePath);
-    UaDeleter<UA_ByteString> privateKeyDeleter(&privateKey, UA_ByteString_deleteMembers);
+    UaDeleter<UA_ByteString> privateKeyDeleter(&privateKey, UA_ByteString_clear);
 
     if (certificate.length == 0) {
         UA_LOG_WARNING(UA_Log_Stdout, UA_LOGCATEGORY_USERLAND,
@@ -154,10 +160,40 @@ bool TestServer::createSecureServerConfig(UA_ServerConfig *config)
         return false;
     }
 
-    result = UA_ServerConfig_addAllSecurityPolicies(config, &certificate, &privateKey);
+    // result = UA_ServerConfig_addAllSecurityPolicies(config, &certificate, &privateKey);
 
-    if (result != UA_STATUSCODE_GOOD) {
-        qWarning() << "Failed to add security policies";
+    // Add the security policies manually because we need to skip Basic128Rsa15 and Basic256
+    // if OpenSSL doesn't support SHA-1 signatures (e.g. RHEL 9).
+
+    UA_StatusCode retval = UA_ServerConfig_addSecurityPolicyNone(config, &certificate);
+    if(retval != UA_STATUSCODE_GOOD) {
+        qWarning() << "Failed to add security policy None";
+        return false;
+    }
+
+    if (Open62541Utils::checkSha1SignatureSupport()) {
+        retval = UA_ServerConfig_addSecurityPolicyBasic128Rsa15(config, &certificate, &privateKey);
+        if(retval != UA_STATUSCODE_GOOD) {
+            qWarning() << "Failed to add security policy Basic128Rsa15";
+            return false;
+        }
+
+        retval = UA_ServerConfig_addSecurityPolicyBasic256(config, &certificate, &privateKey);
+        if(retval != UA_STATUSCODE_GOOD) {
+            qWarning() << "Failed to add security policy Basic256";
+            return false;
+        }
+    }
+
+    retval = UA_ServerConfig_addSecurityPolicyBasic256Sha256(config, &certificate, &privateKey);
+    if(retval != UA_STATUSCODE_GOOD) {
+        qWarning() << "Failed to add security policy Basic256Sha256";
+        return false;
+    }
+
+    retval = UA_ServerConfig_addSecurityPolicyAes128Sha256RsaOaep(config, &certificate, &privateKey);
+    if(retval != UA_STATUSCODE_GOOD) {
+        qWarning() << "Failed to add security policy Aes128Sha256RsaOaep";
         return false;
     }
 
@@ -166,7 +202,7 @@ bool TestServer::createSecureServerConfig(UA_ServerConfig *config)
     certificateDeleter.release();
     privateKeyDeleter.release();
 
-    result = UA_AccessControl_default(config, true,
+    result = UA_AccessControl_default(config, true, nullptr,
                 &config->securityPolicies[0].policyUri,
                 usernamePasswordsSize, usernamePasswords);
 
@@ -204,6 +240,8 @@ bool TestServer::init()
     success = createInsecureServerConfig(m_config);
 #endif
 
+    m_config->maxReferencesPerNode = 10;
+
     m_gathering = UA_HistoryDataGathering_Default(1);
     m_config->historyDatabase = UA_HistoryDatabase_default(m_gathering);
 
@@ -243,9 +281,9 @@ UA_NodeId TestServer::addFolder(const QString &nodeString, const QString &displa
                                      nullptr,
                                      &resultNode);
 
-    UA_QualifiedName_deleteMembers(&nodeBrowseName);
-    UA_NodeId_deleteMembers(&requestedNodeId);
-    UA_ObjectAttributes_deleteMembers(&oAttr);
+    UA_QualifiedName_clear(&nodeBrowseName);
+    UA_NodeId_clear(&requestedNodeId);
+    UA_ObjectAttributes_clear(&oAttr);
 
     if (result != UA_STATUSCODE_GOOD) {
         qWarning() << "Could not add folder:" << nodeString << " :" << result;
@@ -274,8 +312,8 @@ UA_NodeId TestServer::addObject(const UA_NodeId &parentFolder, int namespaceInde
                                      nullptr,
                                      &resultNode);
 
-    UA_QualifiedName_deleteMembers(&nodeBrowseName);
-    UA_ObjectAttributes_deleteMembers(&oAttr);
+    UA_QualifiedName_clear(&nodeBrowseName);
+    UA_ObjectAttributes_clear(&oAttr);
 
     if (result != UA_STATUSCODE_GOOD) {
         qWarning() << "Could not add object to folder:" << result;
@@ -315,8 +353,8 @@ UA_NodeId TestServer::addVariableWithWriteMask(const UA_NodeId &folder, const QS
                                                      &resultId);
 
 
-    UA_NodeId_deleteMembers(&variableNodeId);
-    UA_VariableAttributes_deleteMembers(&attr);
+    UA_NodeId_clear(&variableNodeId);
+    UA_VariableAttributes_clear(&attr);
 
     if (result != UA_STATUSCODE_GOOD) {
         qWarning() << "Could not add variable:" << result;
@@ -366,7 +404,7 @@ UA_NodeId TestServer::addVariable(const UA_NodeId &folder, const QString &variab
                                                      nullptr,
                                                      &resultId);
 
-    // Prevent deletion of the QList's value by UA_VariableAttribute_deleteMembers
+    // Prevent deletion of the QList's value by UA_VariableAttribute_clear
     attr.arrayDimensions = nullptr;
     attr.arrayDimensionsSize = 0;
 
@@ -382,8 +420,8 @@ UA_NodeId TestServer::addVariable(const UA_NodeId &folder, const QString &variab
         }
     }
 
-    UA_NodeId_deleteMembers(&variableNodeId);
-    UA_VariableAttributes_deleteMembers(&attr);
+    UA_NodeId_clear(&variableNodeId);
+    UA_VariableAttributes_clear(&attr);
 
     if (result != UA_STATUSCODE_GOOD) {
         qWarning() << "Could not add variable:" << result << "for node" << variableNode;
@@ -427,8 +465,8 @@ UA_NodeId TestServer::addEmptyArrayVariable(const UA_NodeId &folder, const QStri
                                                      nullptr,
                                                      &resultId);
 
-    UA_NodeId_deleteMembers(&variableNodeId);
-    UA_VariableAttributes_deleteMembers(&attr);
+    UA_NodeId_clear(&variableNodeId);
+    UA_VariableAttributes_clear(&attr);
 
     if (result != UA_STATUSCODE_GOOD) {
         qWarning() << "Could not add empty array variable:" << result;
@@ -507,12 +545,12 @@ UA_NodeId TestServer::addMultiplyMethod(const UA_NodeId &folder, const QString &
                                                      1, &outputArgument,
                                                      nullptr, &resultId);
 
-    UA_QualifiedName_deleteMembers(&nodeBrowseName);
-    UA_NodeId_deleteMembers(&methodNodeId);
-    UA_MethodAttributes_deleteMembers(&attr);
-    UA_Argument_deleteMembers(&inputArguments[0]);
-    UA_Argument_deleteMembers(&inputArguments[1]);
-    UA_Argument_deleteMembers(&outputArgument);
+    UA_QualifiedName_clear(&nodeBrowseName);
+    UA_NodeId_clear(&methodNodeId);
+    UA_MethodAttributes_clear(&attr);
+    UA_Argument_clear(&inputArguments[0]);
+    UA_Argument_clear(&inputArguments[1]);
+    UA_Argument_clear(&outputArgument);
 
     if (result != UA_STATUSCODE_GOOD) {
         qWarning() << "Could not add variable:" << result;
@@ -598,13 +636,13 @@ UA_NodeId TestServer::addMultipleOutputArgumentsMethod(const UA_NodeId &folder,
                                                      2, outputArgument,
                                                      nullptr, &resultId);
 
-    UA_QualifiedName_deleteMembers(&nodeBrowseName);
-    UA_NodeId_deleteMembers(&methodNodeId);
-    UA_MethodAttributes_deleteMembers(&attr);
-    UA_Argument_deleteMembers(&inputArguments[0]);
-    UA_Argument_deleteMembers(&inputArguments[1]);
-    UA_Argument_deleteMembers(&outputArgument[0]);
-    UA_Argument_deleteMembers(&outputArgument[1]);
+    UA_QualifiedName_clear(&nodeBrowseName);
+    UA_NodeId_clear(&methodNodeId);
+    UA_MethodAttributes_clear(&attr);
+    UA_Argument_clear(&inputArguments[0]);
+    UA_Argument_clear(&inputArguments[1]);
+    UA_Argument_clear(&outputArgument[0]);
+    UA_Argument_clear(&outputArgument[1]);
 
     if (result != UA_STATUSCODE_GOOD) {
         qWarning() << "Could not add variable:" << result;
@@ -650,11 +688,11 @@ UA_NodeId TestServer::addAddNamespaceMethod(const UA_NodeId &folder, const QStri
                                                      1, &outputArgument,
                                                      nullptr, &resultId);
 
-    UA_QualifiedName_deleteMembers(&nodeBrowseName);
-    UA_NodeId_deleteMembers(&methodNodeId);
-    UA_MethodAttributes_deleteMembers(&attr);
-    UA_Argument_deleteMembers(&inputArguments[0]);
-    UA_Argument_deleteMembers(&outputArgument);
+    UA_QualifiedName_clear(&nodeBrowseName);
+    UA_NodeId_clear(&methodNodeId);
+    UA_MethodAttributes_clear(&attr);
+    UA_Argument_clear(&inputArguments[0]);
+    UA_Argument_clear(&outputArgument);
 
     if (result != UA_STATUSCODE_GOOD) {
         qWarning() << "Could not add method:" << result;
@@ -721,6 +759,172 @@ UA_NodeId TestServer::addNodeWithFixedTimestamp(const UA_NodeId &folder, const Q
     return resultId;
 }
 
+UA_StatusCode TestServer::addServerStatusTypeTestNodes(const UA_NodeId &parent)
+{
+    const auto createTestValue = [](int index) -> UA_ServerStatusDataType {
+        UA_ServerStatusDataType s;
+        UA_ServerStatusDataType_init(&s);
+
+        const UA_DateTime testDateTime = UA_DATETIME_UNIX_EPOCH + 1691996809123 * UA_DATETIME_MSEC;
+
+        s.currentTime = testDateTime + index * UA_DATETIME_MSEC;
+        s.startTime = testDateTime + index * UA_DATETIME_MSEC;
+        s.secondsTillShutdown = 23;
+        s.shutdownReason = UA_LOCALIZEDTEXT_ALLOC("en", "ShutdownReason");
+        s.state = UA_SERVERSTATE_TEST;
+        s.buildInfo.buildDate = testDateTime + index * UA_DATETIME_MSEC;
+        s.buildInfo.buildNumber = UA_STRING_ALLOC("BuildNumber");
+        s.buildInfo.manufacturerName = UA_STRING_ALLOC("ManufacturerName");
+        s.buildInfo.productName = UA_STRING_ALLOC("ProductName");
+        s.buildInfo.productUri = UA_STRING_ALLOC("ProductUri");
+        s.buildInfo.softwareVersion = UA_STRING_ALLOC("SoftwareVersion");
+
+        return s;
+    };
+
+    {
+        UA_NodeId variableNodeId = Open62541Utils::nodeIdFromQString("ns=3;s=ServerStatusScalar");
+
+        auto value = UA_ServerStatusDataType_new();
+        *value = createTestValue(0);
+        auto var = UA_Variant();
+        UA_Variant_init(&var);
+        UA_Variant_setScalar(&var, value, &UA_TYPES[UA_TYPES_SERVERSTATUSDATATYPE]);
+
+        UA_VariableAttributes attr = UA_VariableAttributes_default;
+        attr.value = var;
+        attr.valueRank = UA_VALUERANK_SCALAR;
+        attr.displayName = UA_LOCALIZEDTEXT_ALLOC("en-US", "ServerStatusScalar");
+        attr.dataType = attr.value.type->typeId;
+        attr.accessLevel = UA_ACCESSLEVELMASK_READ | UA_ACCESSLEVELMASK_WRITE;
+
+        UA_QualifiedName variableName;
+        variableName.namespaceIndex = variableNodeId.namespaceIndex;
+        variableName.name = UA_STRING_STATIC("ServerStatusScalar");
+
+        UA_NodeId resultId;
+        UA_StatusCode result = UA_Server_addVariableNode(m_server,
+                                                         variableNodeId,
+                                                         parent,
+                                                         UA_NODEID_NUMERIC(0, UA_NS0ID_ORGANIZES),
+                                                         variableName,
+                                                         UA_NODEID_NULL,
+                                                         attr,
+                                                         nullptr,
+                                                         &resultId);
+
+        UA_NodeId_clear(&variableNodeId);
+        UA_VariableAttributes_clear(&attr);
+
+        if (result != UA_STATUSCODE_GOOD) {
+            qWarning() << "Could not add scalar server status variable";
+            return result;
+        }
+    }
+
+    {
+        UA_NodeId variableNodeId = Open62541Utils::nodeIdFromQString("ns=3;s=ServerStatusArray");
+
+        auto value = static_cast<UA_ServerStatusDataType *>(UA_Array_new(2, &UA_TYPES[UA_TYPES_SERVERSTATUSDATATYPE]));
+        for (int i = 0; i < 2; ++i)
+            value[i] = createTestValue(i);
+        auto var = UA_Variant();
+        UA_Variant_init(&var);
+        UA_Variant_setArray(&var, value, 2, &UA_TYPES[UA_TYPES_SERVERSTATUSDATATYPE]);
+
+        UA_VariableAttributes attr = UA_VariableAttributes_default;
+        attr.value = var;
+        attr.valueRank = UA_VALUERANK_ONE_DIMENSION;
+        quint32 arrayDimensions[] = { 0 };
+        attr.arrayDimensionsSize = 1;
+        attr.arrayDimensions = arrayDimensions;
+        attr.displayName = UA_LOCALIZEDTEXT_ALLOC("en-US", "ServerStatusArray");
+        attr.dataType = attr.value.type->typeId;
+        attr.accessLevel = UA_ACCESSLEVELMASK_READ | UA_ACCESSLEVELMASK_WRITE;
+
+        UA_QualifiedName variableName;
+        variableName.namespaceIndex = variableNodeId.namespaceIndex;
+        variableName.name = UA_STRING_STATIC("ServerStatusArray");
+
+        UA_NodeId resultId;
+        UA_StatusCode result = UA_Server_addVariableNode(m_server,
+                                                         variableNodeId,
+                                                         parent,
+                                                         UA_NODEID_NUMERIC(0, UA_NS0ID_ORGANIZES),
+                                                         variableName,
+                                                         UA_NODEID_NULL,
+                                                         attr,
+                                                         nullptr,
+                                                         &resultId);
+
+        attr.arrayDimensionsSize = 0;
+        attr.arrayDimensions = nullptr;
+
+        UA_NodeId_clear(&variableNodeId);
+        UA_VariableAttributes_clear(&attr);
+
+        if (result != UA_STATUSCODE_GOOD) {
+            qWarning() << "Could not add array server status variable:" << UA_StatusCode_name(result);
+            return result;
+        }
+    }
+
+    {
+        UA_NodeId variableNodeId = Open62541Utils::nodeIdFromQString("ns=3;s=ServerStatusMultiDimensionalArray");
+
+        auto value = static_cast<UA_ServerStatusDataType *>(UA_Array_new(4, &UA_TYPES[UA_TYPES_SERVERSTATUSDATATYPE]));
+        for (int i = 0; i < 4; ++i)
+            value[i] = createTestValue(i);
+        auto var = UA_Variant();
+        UA_Variant_init(&var);
+        UA_Variant_setArray(&var, value, 4, &UA_TYPES[UA_TYPES_SERVERSTATUSDATATYPE]);
+
+        quint32 arrayDimensions[] = { 2, 2 };
+
+        var.arrayDimensionsSize = 2;
+        var.arrayDimensions = arrayDimensions;
+
+        UA_VariableAttributes attr = UA_VariableAttributes_default;
+        attr.value = var;
+        attr.valueRank = UA_VALUERANK_TWO_DIMENSIONS;
+        attr.arrayDimensionsSize = 2;
+        attr.arrayDimensions = arrayDimensions;
+        attr.displayName = UA_LOCALIZEDTEXT_ALLOC("en-US", "ServerStatusMultiDimensionalArray");
+        attr.dataType = attr.value.type->typeId;
+        attr.accessLevel = UA_ACCESSLEVELMASK_READ | UA_ACCESSLEVELMASK_WRITE;
+
+        UA_QualifiedName variableName;
+        variableName.namespaceIndex = variableNodeId.namespaceIndex;
+        variableName.name = UA_STRING_STATIC("ServerStatusMultiDimensionalArray");
+
+        UA_NodeId resultId;
+        UA_StatusCode result = UA_Server_addVariableNode(m_server,
+                                                         variableNodeId,
+                                                         parent,
+                                                         UA_NODEID_NUMERIC(0, UA_NS0ID_ORGANIZES),
+                                                         variableName,
+                                                         UA_NODEID_NULL,
+                                                         attr,
+                                                         nullptr,
+                                                         &resultId);
+
+        attr.arrayDimensionsSize = 0;
+        attr.arrayDimensions = nullptr;
+        attr.value.arrayDimensionsSize = 0;
+        attr.value.arrayDimensions = nullptr;
+
+        UA_NodeId_clear(&variableNodeId);
+        UA_VariableAttributes_clear(&attr);
+
+        if (result != UA_STATUSCODE_GOOD) {
+            qWarning() << "Could not add multi dimensional array server status variable:" << UA_StatusCode_name(result);
+            return result;
+        }
+    }
+
+    return UA_STATUSCODE_GOOD;
+}
+
 // The event test methods are based on the open62541 tutorial_server_events.c example
 UA_StatusCode TestServer::generateEventCallback(UA_Server *server,
                                                 const UA_NodeId *sessionId, void *sessionHandle,
@@ -844,4 +1048,186 @@ UA_StatusCode TestServer::addEventTrigger(const UA_NodeId &parent)
     return result;
 }
 
+static UA_DataTypeArray customTypes = {
+    nullptr,
+    UA_TYPES_QTOPCUATESTMODEL_COUNT,
+    &UA_TYPES_QTOPCUATESTMODEL[0]
+};
+
+// This method must be called after the other test namespaces have been added
+UA_StatusCode TestServer::addEncoderTestModel()
+{
+    auto result = namespace_qtopcuatestmodel_generated(m_server);
+
+    if (result != UA_STATUSCODE_GOOD)
+        return result;
+
+    auto config = UA_Server_getConfig(m_server);
+    customTypes.next = config->customDataTypes;
+    config->customDataTypes = &customTypes;
+
+    {
+        UA_QtTestStructType data;
+        UA_QtTestStructType_init(&data);
+
+        data.enumMember = UA_QTTESTENUMERATION_FIRSTOPTION;
+        data.int64ArrayMemberSize = 3;
+        data.int64ArrayMember = static_cast<UA_Int64 *>(UA_Array_new(3, &UA_TYPES[UA_TYPES_INT64]));
+        data.int64ArrayMember[0] = std::numeric_limits<qint64>::max();
+        data.int64ArrayMember[1] = std::numeric_limits<qint64>::max() - 1;
+        data.int64ArrayMember[2] = std::numeric_limits<qint64>::min();
+
+        data.stringMember = UA_STRING_ALLOC("TestString");
+        data.localizedTextMember = UA_LOCALIZEDTEXT_ALLOC("en", "TestText");
+        data.qualifiedNameMember = UA_QUALIFIEDNAME_ALLOC(1, "TestName");
+
+        data.nestedStructMember.doubleSubtypeMember = 42;
+
+        data.nestedStructArrayMemberSize = 2;
+        data.nestedStructArrayMember = static_cast<UA_QtInnerTestStructType *>(
+            UA_Array_new(2, &UA_TYPES_QTOPCUATESTMODEL[UA_TYPES_QTOPCUATESTMODEL_QTINNERTESTSTRUCTTYPE]));
+
+        data.nestedStructArrayMember[0].doubleSubtypeMember = 23.0;
+        data.nestedStructArrayMember[1].doubleSubtypeMember = 42.0;
+
+        UA_Variant var;
+        UA_Variant_init(&var);
+        UA_Variant_setScalarCopy(&var, &data, &UA_TYPES_QTOPCUATESTMODEL[UA_TYPES_QTOPCUATESTMODEL_QTTESTSTRUCTTYPE]);
+
+        result = UA_Server_writeValue(m_server, UA_NODEID_NUMERIC(4, UA_QTOPCUATESTMODELID_DECODERTESTNODES_NESTEDSTRUCT), var);
+        UA_Variant_clear(&var);
+
+        if (result != UA_STATUSCODE_GOOD) {
+            qWarning() << "Failed to write nested struct node";
+            return result;
+        }
+    }
+
+    {
+        UA_QtTestUnionType data;
+        UA_QtTestUnionType_init(&data);
+
+        data.switchField = UA_QTTESTUNIONTYPESWITCH_MEMBER1;
+        data.fields.member1 = 42;
+
+        UA_Variant var;
+        UA_Variant_init(&var);
+        UA_Variant_setScalarCopy(&var, &data, &UA_TYPES_QTOPCUATESTMODEL[UA_TYPES_QTOPCUATESTMODEL_QTTESTUNIONTYPE]);
+
+        result = UA_Server_writeValue(m_server, UA_NODEID_NUMERIC(4, UA_QTOPCUATESTMODELID_DECODERTESTNODES_UNIONWITHFIRSTMEMBER), var);
+        UA_Variant_clear(&var);
+
+        if (result != UA_STATUSCODE_GOOD) {
+            qWarning() << "Failed to write union struct node";
+            return result;
+        }
+    }
+
+    {
+        UA_QtTestUnionType data;
+        UA_QtTestUnionType_init(&data);
+
+        data.switchField = UA_QTTESTUNIONTYPESWITCH_MEMBER2;
+        data.fields.member2.doubleSubtypeMember = 23.0;
+
+        UA_Variant var;
+        UA_Variant_init(&var);
+        UA_Variant_setScalarCopy(&var, &data, &UA_TYPES_QTOPCUATESTMODEL[UA_TYPES_QTOPCUATESTMODEL_QTTESTUNIONTYPE]);
+
+        result = UA_Server_writeValue(m_server, UA_NODEID_NUMERIC(4, UA_QTOPCUATESTMODELID_DECODERTESTNODES_UNIONWITHSECONDMEMBER), var);
+        UA_Variant_clear(&var);
+
+        if (result != UA_STATUSCODE_GOOD) {
+            qWarning() << "Failed to write union struct node";
+            return result;
+        }
+    }
+
+    {
+        UA_QtStructWithOptionalFieldType data;
+        UA_QtStructWithOptionalFieldType_init(&data);
+
+        data.mandatoryMember =  42.0;
+        data.optionalMember = UA_Double_new();
+        *data.optionalMember = 23.0;
+
+        UA_Variant var;
+        UA_Variant_init(&var);
+        UA_Variant_setScalarCopy(&var, &data, &UA_TYPES_QTOPCUATESTMODEL[UA_TYPES_QTOPCUATESTMODEL_QTSTRUCTWITHOPTIONALFIELDTYPE]);
+
+        result = UA_Server_writeValue(m_server, UA_NODEID_NUMERIC(4, UA_QTOPCUATESTMODELID_DECODERTESTNODES_STRUCTWITHOPTIONALFIELD), var);
+        UA_Variant_clear(&var);
+
+        if (result != UA_STATUSCODE_GOOD) {
+            qWarning() << "Failed to write optional field struct node";
+            return result;
+        }
+    }
+
+    {
+        UA_QtStructWithOptionalFieldType data;
+        UA_QtStructWithOptionalFieldType_init(&data);
+
+        data.mandatoryMember =  42.0;
+
+        UA_Variant var;
+        UA_Variant_init(&var);
+        UA_Variant_setScalarCopy(&var, &data, &UA_TYPES_QTOPCUATESTMODEL[UA_TYPES_QTOPCUATESTMODEL_QTSTRUCTWITHOPTIONALFIELDTYPE]);
+
+        result = UA_Server_writeValue(m_server, UA_NODEID_NUMERIC(4, UA_QTOPCUATESTMODELID_DECODERTESTNODES_STRUCTWIHOUTOPTIONALFIELD), var);
+        UA_Variant_clear(&var);
+
+        if (result != UA_STATUSCODE_GOOD) {
+            qWarning() << "Failed to write optional field struct node";
+            return result;
+        }
+    }
+
+    {
+        auto data = UA_QtTestStructWithDiagnosticInfo_new();
+
+        data->diagnosticInfoMember.hasSymbolicId = true;
+        data->diagnosticInfoMember.symbolicId = 1;
+        data->diagnosticInfoMember.hasNamespaceUri = true;
+        data->diagnosticInfoMember.namespaceUri = 2;
+        data->diagnosticInfoMember.hasLocalizedText = true;
+        data->diagnosticInfoMember.localizedText = 3;
+        data->diagnosticInfoMember.hasLocale = true;
+        data->diagnosticInfoMember.locale = 4;
+        data->diagnosticInfoMember.hasAdditionalInfo = true;
+        data->diagnosticInfoMember.additionalInfo = UA_STRING_ALLOC("My additional info");
+        data->diagnosticInfoMember.hasInnerStatusCode = true;
+        data->diagnosticInfoMember.innerStatusCode = UA_STATUSCODE_BADINTERNALERROR;
+        data->diagnosticInfoMember.hasInnerDiagnosticInfo = true;
+
+        data->diagnosticInfoMember.innerDiagnosticInfo = UA_DiagnosticInfo_new();
+        data->diagnosticInfoMember.innerDiagnosticInfo->hasAdditionalInfo = true;
+        data->diagnosticInfoMember.innerDiagnosticInfo->additionalInfo = UA_STRING_ALLOC("My inner additional info");
+
+        data->diagnosticInfoArrayMemberSize = 2;
+        data->diagnosticInfoArrayMember = static_cast<UA_DiagnosticInfo *>(UA_Array_new(2, &UA_TYPES[UA_TYPES_DIAGNOSTICINFO]));
+
+        UA_DiagnosticInfo_copy(&data->diagnosticInfoMember, &data->diagnosticInfoArrayMember[0]);
+
+        data->diagnosticInfoArrayMember[1].hasLocale = true;
+        data->diagnosticInfoArrayMember[1].locale = 1;
+        data->diagnosticInfoArrayMember[1].hasInnerStatusCode = true;
+        data->diagnosticInfoArrayMember[1].innerStatusCode = UA_STATUSCODE_BADTYPEMISMATCH;
+
+        UA_Variant var;
+        UA_Variant_init(&var);
+        UA_Variant_setScalar(&var, data, &UA_TYPES_QTOPCUATESTMODEL[UA_TYPES_QTOPCUATESTMODEL_QTTESTSTRUCTWITHDIAGNOSTICINFO]);
+
+        result = UA_Server_writeValue(m_server, UA_NODEID_NUMERIC(4, UA_QTOPCUATESTMODELID_DECODERTESTNODES_STRUCTWITHDIAGNOSTICINFO), var);
+        UA_Variant_clear(&var);
+
+        if (result != UA_STATUSCODE_GOOD) {
+            qWarning() << "Failed to write diagnostic info field struct node";
+            return result;
+        }
+    }
+
+    return result;
+}
+
 QT_END_NAMESPACE
